@@ -35,11 +35,10 @@
 #endif
 
 #include <gio/gio.h>
+#include <libxfce4ui/libxfce4ui.h>
 #include <exo-desktop-item-edit/exo-die-editor.h>
 #include <exo-desktop-item-edit/exo-die-utils.h>
 
-#include <exo-support/exo-support.h>
-
 #if defined(GDK_WINDOWING_X11)
 #include <gdk/gdkx.h>
 #endif
@@ -65,7 +64,7 @@ static gchar   *opt_comment = NULL;
 static gchar   *opt_command = NULL;
 static gchar   *opt_url = NULL;
 static gchar   *opt_icon = NULL;
-static gint     opt_xid = 0;
+static gint64   opt_xid = 0;
 
 
 
@@ -84,7 +83,7 @@ static GOptionEntry option_entries[] =
   { "url", 0, 0, G_OPTION_ARG_STRING, &opt_url, N_ ("Preset URL when creating a link"), NULL, },
   { "icon", 0, 0, G_OPTION_ARG_STRING, &opt_icon, N_ ("Preset icon when creating a desktop file"), NULL, },
   { "version", 'V', 0, G_OPTION_ARG_NONE, &opt_version, N_ ("Print version information and exit"), NULL, },
-  { "xid", 0, G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_INT, &opt_xid, NULL, NULL, },
+  { "xid", 0, G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_INT64, &opt_xid, NULL, NULL, },
   { NULL, },
 };
 
@@ -134,6 +133,13 @@ main (int argc, char **argv)
   gsize            length = 0;
   gboolean         res;
   GFileType        file_type;
+  GFile           *gfile_local;
+  gchar           *relpath;
+  gchar           *path;
+  gchar          **dirs;
+  guint            i;
+  const gchar     *mode_dir;
+  gint             ox, oy, ow, oh;
 
   /* setup translation domain */
   xfce_textdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR, "UTF-8");
@@ -398,7 +404,7 @@ main (int argc, char **argv)
   if (G_UNLIKELY (opt_xid != 0))
     {
       /* try to determine the window for the id */
-      xwindow = gdk_window_foreign_new (opt_xid);
+      xwindow = gdk_window_foreign_new ((GdkNativeWindow) opt_xid);
       if (G_LIKELY (xwindow != NULL))
         {
           /* realize the dialog first... */
@@ -406,6 +412,16 @@ main (int argc, char **argv)
 
           /* ...and set the "transient for" relation */
           gdk_window_set_transient_for (dialog->window, xwindow);
+          gtk_window_set_screen (GTK_WINDOW (dialog), gdk_window_get_screen (xwindow));
+
+          /* center on parent */
+          gdk_window_get_root_origin (xwindow, &ox, &oy);
+          gdk_window_get_geometry (xwindow, NULL, NULL, &ow, &oh, NULL);
+
+          ox += (ow - dialog->allocation.width) / 2;
+          oy += (oh - dialog->allocation.height) / 2;
+
+          gtk_window_move (GTK_WINDOW (dialog), MAX (ox, 0), MAX (oy, 0));
         }
     }
 #endif
@@ -465,59 +481,108 @@ main (int argc, char **argv)
         }
 
       /* try to save the file */
-      if (!exo_die_g_key_file_save (key_file, opt_create_new, gfile, mode, &error)
-          && opt_create_new)
+      if (!exo_die_g_key_file_save (key_file, opt_create_new, gfile, mode, &error))
         {
-          /* reset the error */
-          g_clear_error (&error);
-
-          /* create failed, ask the user to specify a file name */
-          chooser = gtk_file_chooser_dialog_new (_("Choose filename"),
-                                                 GTK_WINDOW (dialog),
-                                                 GTK_FILE_CHOOSER_ACTION_SAVE,
-                                                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-                                                 GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
-                                                 NULL);
-          gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (chooser), TRUE);
-          gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (chooser), TRUE);
-
-          file_type = g_file_query_file_type (gfile, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL);
-
-          /* if base is a folder, enter the folder */
-          if (file_type == G_FILE_TYPE_DIRECTORY)
+          if (opt_create_new)
             {
-              gtk_file_chooser_set_current_folder_file (GTK_FILE_CHOOSER (chooser), gfile, NULL);
-              gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (chooser), "new-file.desktop");
-            }
-          else if (file_type == G_FILE_TYPE_REGULAR)
-            {
-              gfile_parent = g_file_get_parent (gfile);
-              if (G_LIKELY (gfile_parent != NULL))
+              /* reset the error */
+              g_clear_error (&error);
+
+              /* create failed, ask the user to specify a file name */
+              chooser = gtk_file_chooser_dialog_new (_("Choose filename"),
+                                                     GTK_WINDOW (dialog),
+                                                     GTK_FILE_CHOOSER_ACTION_SAVE,
+                                                     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                                     GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
+                                                     NULL);
+              gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (chooser), TRUE);
+              gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (chooser), TRUE);
+
+              file_type = g_file_query_file_type (gfile, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL);
+
+              /* if base is a folder, enter the folder */
+              if (file_type == G_FILE_TYPE_DIRECTORY)
                 {
-                  gtk_file_chooser_set_current_folder_file (GTK_FILE_CHOOSER (chooser), gfile_parent, NULL);
-                  g_object_unref (G_OBJECT (gfile_parent));
+                  gtk_file_chooser_set_current_folder_file (GTK_FILE_CHOOSER (chooser), gfile, NULL);
+                  gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (chooser), "new-file.desktop");
+                }
+              else if (file_type == G_FILE_TYPE_REGULAR)
+                {
+                  gfile_parent = g_file_get_parent (gfile);
+                  if (G_LIKELY (gfile_parent != NULL))
+                    {
+                      gtk_file_chooser_set_current_folder_file (GTK_FILE_CHOOSER (chooser), gfile_parent, NULL);
+                      g_object_unref (G_OBJECT (gfile_parent));
+                    }
+
+                  base_name = g_file_get_basename (gfile);
+                  gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (chooser), base_name);
+                  g_free (base_name);
                 }
 
-              base_name = g_file_get_basename (gfile);
-              gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (chooser), base_name);
-              g_free (base_name);
-            }
+              /* run the chooser */
+              response = gtk_dialog_run (GTK_DIALOG (chooser));
+              if (G_LIKELY (response == GTK_RESPONSE_ACCEPT))
+                {
+                  /* release the previous file name */
+                  if (G_LIKELY (gfile != NULL))
+                    g_object_unref (G_OBJECT (gfile));
 
-          /* run the chooser */
-          response = gtk_dialog_run (GTK_DIALOG (chooser));
-          if (G_LIKELY (response == GTK_RESPONSE_ACCEPT))
+                  /* try again to save to the new file */
+                  gfile = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (chooser));
+                  exo_die_g_key_file_save (key_file, FALSE, gfile, mode, &error);
+                }
+
+              /* destroy the chooser */
+              gtk_widget_destroy (chooser);
+            }
+          else if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
             {
-              /* release the previous file name */
-              if (G_LIKELY (gfile != NULL))
-                g_object_unref (G_OBJECT (gfile));
+              if (mode == EXO_DIE_EDITOR_MODE_DIRECTORY)
+                mode_dir = "desktop-directories/";
+              else
+                mode_dir = "applications/";
 
-              /* try again to save to the new file */
-              gfile = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (chooser));
-              exo_die_g_key_file_save (key_file, FALSE, gfile, mode, &error);
+              /* check if the file is in one of the applications directories
+               * and get the relative path */
+              dirs = xfce_resource_lookup_all (XFCE_RESOURCE_DATA, mode_dir);
+              base_name = NULL;
+              for (base_name = NULL, i = 0; !base_name && dirs[i] != NULL; i++)
+                {
+                  gfile_parent = g_file_new_for_path (dirs[i]);
+                  base_name = g_file_get_relative_path (gfile_parent, gfile);
+                  g_object_unref (G_OBJECT (gfile_parent));
+                }
+              g_strfreev (dirs);
+
+              /* file was found in an applications directory, write a new file in
+               * the users' local applications directory */
+              if (base_name != NULL)
+                {
+                  /* get the new file location */
+                  relpath = g_build_filename (mode_dir, base_name, NULL);
+                  path = xfce_resource_save_location (XFCE_RESOURCE_DATA, relpath, TRUE);
+                  g_free (relpath);
+
+                  if (G_LIKELY (path != NULL))
+                    {
+                      /* silently notify the user we're going to write to a new location */
+                      exo_die_error ("\"%s\" is not writeable, saving to \"%s\" instead.",
+                                     argv[1], path);
+
+                      /* reset the error */
+                      g_clear_error (&error);
+
+                      /* try another save */
+                      gfile_local = g_file_new_for_path (path);
+                      exo_die_g_key_file_save (key_file, FALSE, gfile_local, mode, &error);
+                      g_object_unref (G_OBJECT (gfile_local));
+                    }
+
+                  g_free (path);
+                  g_free (base_name);
+                }
             }
-
-          /* destroy the chooser */
-          gtk_widget_destroy (chooser);
         }
 
       /* check if we failed to save/create */
